import boto3
import botocore.exceptions
import os
import json
import cfnresponse
from typing import Dict
from enum import Enum


class Operation(Enum):
    UPDATE_BUILDKIT_ISRA_ROLE = 'update-buildkit-isra-role'
    EMPTY_S3_BUCKET = 'empty-s3-bucket'
    DELETE_KARPENTER_NODE_ROLE_AND_INSTANCE_PROFILE = 'delete-karpenter-node-role-and-instance-profile'
    DELETE_ECR_REPO_IMAGES = 'delete-ecr-repo-images'
    TEARDOWN_ALL_VOLUMES = 'teardown-all-volumes'


def get_aws_caller_identity():
    return boto3.client('sts').get_caller_identity()

def get_cluster_info(cluster_name: str):
    eks = boto3.client('eks')
    cluster = eks.describe_cluster(name=cluster_name)['cluster']
    return {
        'endpoint': cluster['endpoint'],
        'ca': cluster['certificateAuthority']['data']
    }


def empty_s3_bucket(bucket_name: str):
    s3 = boto3.client('s3')
    paginator = s3.get_paginator('list_objects_v2')
    objects_deleted = 0
    for page in paginator.paginate(Bucket=bucket_name):
        if 'Contents' in page:
            for obj in page['Contents']:
                s3.delete_object(Bucket=bucket_name, Key=obj['Key'])
                objects_deleted += 1
    print(f"Deleted {objects_deleted} objects from bucket {bucket_name}")
    return True


def update_buildkit_isra_role_trust_policy(cluster_name: str, oidc_provider_url: str, aws_account_id: str,
                                           service_account_name: str, namespace: str):
    iam = boto3.client('iam')
    role_name = f"{cluster_name}-buildkit-ecr-role"
    response = iam.get_role(RoleName=role_name)

    oidc_issuer = oidc_provider_url.split('//')[1]

    current_trust_policy = response['Role']['AssumeRolePolicyDocument']
    current_sub_values = current_trust_policy['Statement'][0]['Condition']['StringEquals'].get(
        "{}:sub".format(oidc_issuer), None)
    if isinstance(current_sub_values, str):
        current_sub_values = [current_sub_values]
    new_service_account = "system:serviceaccount:{}:{}".format(namespace, service_account_name)
    if new_service_account not in current_sub_values:
        current_sub_values.append(new_service_account)
        trust_policy = current_trust_policy
        trust_policy['Statement'][0]['Condition']['StringEquals']["{}:sub".format(oidc_issuer)] = current_sub_values

        # Update the trust policy with the new service account
        iam.update_assume_role_policy(RoleName=role_name, PolicyDocument=json.dumps(trust_policy))
        print(f"IAM role {role_name} trust policy updated successfully.")
    else:
        print(f"Service account {new_service_account} already exists in the IAM role trust policy.")
    return True


def delete_role_and_attached_policies(iam, role_name, cluster_name):
    # List all attached policies
    attached_policies = iam.list_attached_role_policies(RoleName=role_name)['AttachedPolicies']

    # Detach each policy
    for policy in attached_policies:
        iam.detach_role_policy(RoleName=role_name, PolicyArn=policy['PolicyArn'])
        if cluster_name in policy['PolicyName']:
            # Delete the policy
            iam.delete_policy(PolicyArn=policy['PolicyArn'])

    # Delete the role
    iam.delete_role(RoleName=role_name)

def delete_karpenter_node_role_and_instance_profile(cluster_name: str):
    iam = boto3.client('iam')
    role_name = f'KarpenterNodeRole-{cluster_name}'

    # List all instance profiles
    instance_profiles = iam.list_instance_profiles()['InstanceProfiles']

    # For each instance profile
    for profile in instance_profiles:
        # Check if the role is associated with the instance profile
        for role in profile['Roles']:
            if role['RoleName'] == role_name:
                # Remove the role from the instance profile
                iam.remove_role_from_instance_profile(InstanceProfileName=profile['InstanceProfileName'],
                    RoleName=role_name)

                # Delete the instance profile
                iam.delete_instance_profile(InstanceProfileName=profile['InstanceProfileName'])

    # Delete the role
    delete_role_and_attached_policies(iam, role_name, cluster_name)
    return True


def delete_ecr_repo_images(repository_name: str):
    ecr = boto3.client('ecr')
    try:
        ecr.describe_repositories(repositoryNames=[repository_name])
    except ecr.exceptions.RepositoryNotFoundException:
        print(f"Repository {repository_name} does not exist.")
        return True

    print(f"Fetching images from ECR for repository: {repository_name}...")
    images = ecr.list_images(repositoryName=repository_name)['imageIds']
    if images:
        print(f"{len(images)} Images found. Deleting images...")
        ecr.batch_delete_image(repositoryName=repository_name, imageIds=images)
    else:
        print("No images found in the repository.")
    return True


def delete_cloudformation_stack(stack_name: str, wait: bool = False):
    cloudformation_client = boto3.client('cloudformation')
    try:
        cloudformation_client.describe_stacks(StackName=stack_name)
    except botocore.exceptions.ClientError as e:
        if 'does not exist' in str(e):
            print(f'Stack {stack_name} does not exist.')
            return True
        else:
            raise

    print(f"Deleting stack: {stack_name}")
    response = cloudformation_client.delete_stack(StackName=stack_name)
    if wait:
        waiter = cloudformation_client.get_waiter('stack_delete_complete')
        waiter.wait(StackName=stack_name, WaiterConfig={'MaxAttempts': 30})
    return response


def delete_all_cluster_efs_volumes(cluster_name: str) -> bool:
    efs_client = boto3.client('efs')
    response = efs_client.describe_file_systems()
    file_systems = response['FileSystems']
    deleted_stacks = []
    try:
        for fs in file_systems:
            tags = efs_client.describe_tags(FileSystemId=fs['FileSystemId'])['Tags']
            created_by = next((tag['Value'] for tag in tags if tag['Key'] == 'CreatedBy'), None)
            efs_cluster_name = next((tag['Value'] for tag in tags if tag['Key'] == 'ClusterName'), None)
            if created_by != 'Tensorfuse' or efs_cluster_name != cluster_name:
                continue

            name = next((tag['Value'] for tag in tags if tag['Key'] == 'Name'), None)

            efs_vol_stack_name = f"{cluster_name}-efs-volume-{name}"
            print(f"Deleting volume: {name}")
            delete_cloudformation_stack(efs_vol_stack_name, wait=False)
            deleted_stacks.append(efs_vol_stack_name)

        cloudformation_client = boto3.client('cloudformation')
        waiter = cloudformation_client.get_waiter('stack_delete_complete')
        for stack in deleted_stacks:
            print(f"Waiting for stack '{stack}' to delete...")
            waiter.wait(StackName=stack, WaiterConfig={'MaxAttempts': 30})
            print("Stack deleted.")
    except Exception as e:
        print(f"Error deleting EFS volumes: {str(e)}")
        return False

    return True


def delete_all_volumes(cluster_name: str):
    print("Deleting EFS Volumes...")
    delete_all_cluster_efs_volumes(cluster_name)
    print("Deleting EFS Volume Support...")
    delete_cloudformation_stack(f"{cluster_name}-volume-support-efs", wait=True)
    return True


def perform_create_operation(operation: str, parameters: Dict) -> bool:
    if operation == Operation.UPDATE_BUILDKIT_ISRA_ROLE.value:
        service_account_name = parameters.get('ServiceAccountName')
        namespace = parameters.get('Namespace')
        cluster_name = parameters.get('ClusterName')
        oidc_provider_url = parameters.get('OIDCProviderUrl')
        aws_account_id = parameters.get('AWSAccountId')
        if not (service_account_name and namespace and cluster_name and oidc_provider_url and aws_account_id):
            raise Exception("RoleArn, ServiceAccountName and Namespace are required parameters")
        update_buildkit_isra_role_trust_policy(cluster_name, oidc_provider_url, aws_account_id, service_account_name, namespace)
        return True
    elif operation in [Operation.EMPTY_S3_BUCKET.value, Operation.DELETE_KARPENTER_NODE_ROLE_AND_INSTANCE_PROFILE.value,
                       Operation.DELETE_ECR_REPO_IMAGES.value, Operation.TEARDOWN_ALL_VOLUMES.value]:
        return True
    else:
        return False


def perform_delete_operation(operation: str, parameters: Dict) -> bool:
    if operation == Operation.EMPTY_S3_BUCKET.value:
        bucket_name = parameters.get('BucketName')
        if not bucket_name:
            raise Exception("BucketName is a required parameter")
        empty_s3_bucket(bucket_name)
        return True
    elif operation == Operation.DELETE_KARPENTER_NODE_ROLE_AND_INSTANCE_PROFILE.value:
        cluster_name = parameters.get('ClusterName')
        if not cluster_name:
            raise Exception("ClusterName is a required parameter")
        delete_karpenter_node_role_and_instance_profile(cluster_name)
        return True
    elif operation == Operation.DELETE_ECR_REPO_IMAGES.value:
        repository_name = parameters.get('RepositoryName')
        if not repository_name:
            raise Exception("RepositoryName is a required parameter")
        delete_ecr_repo_images(repository_name)
        return True
    elif operation == Operation.TEARDOWN_ALL_VOLUMES.value:
        cluster_name = parameters.get('ClusterName')
        if not cluster_name:
            raise Exception("ClusterName is a required parameter")
        return delete_all_volumes(cluster_name)
    elif operation in [Operation.UPDATE_BUILDKIT_ISRA_ROLE.value]:
        return True
    else:
        return False

def handler(event, context):
    command = event['ResourceProperties'].get('Operation')
    parameters = event['ResourceProperties'].get('Parameters')

    os.environ['XDG_CONFIG_HOME'] = '/tmp/.config'
    os.environ['XDG_CACHE_HOME'] = '/tmp/.cache'

    if event['RequestType'] == 'Create':
        try:
            success = perform_create_operation(command, parameters)

            if success:
                cfnresponse.send(event, context, cfnresponse.SUCCESS, {'Event': 'Create', 'Reason': 'Operation successful'})
            else:
                cfnresponse.send(event, context, cfnresponse.FAILED, {'Event': 'Create', 'Reason': 'Invalid operation or parameters'})
        except Exception as e:
            cfnresponse.send(event, context, cfnresponse.FAILED, {'Event': 'Create', 'Reason': str(e)})
    elif event['RequestType'] == 'Update':
        cfnresponse.send(event, context, cfnresponse.SUCCESS, {'Event': 'Update', 'Reason': 'No action needed'})
    elif event['RequestType'] == 'Delete':
        try:
            success = perform_delete_operation(command, parameters)

            if success:
                cfnresponse.send(event, context, cfnresponse.SUCCESS, {'Event': 'Delete', 'Reason': 'Operation successful'})
            else:
                cfnresponse.send(event, context, cfnresponse.FAILED, {'Event': 'Delete', 'Reason': 'Invalid operation or parameters'})
        except Exception as e:
            cfnresponse.send(event, context, cfnresponse.FAILED, {'Event': 'Delete', 'Reason': str(e)})
    else:
        cfnresponse.send(event, context, cfnresponse.FAILED, {'Reason': 'Invalid request type'})



from typing import Dict
import re

from tensorkube.services.filesystem_service import create_efs_volume_k8s_resources, get_efs_filesystem_by_name, \
    check_existing_efs_volume_k8s_resources

SUPPORTED_VOLUME_TYPES = ["efs"]

def volume_name_satisfies_regex(volume_name: str):
    pattern = r'^[a-z0-9]([-a-z0-9]*[a-z0-9])?$'
    return bool(re.match(pattern, volume_name))

def get_deployment_volume_name(volume_name: str, volume_type: str):
    return f"{volume_type}-{volume_name}"

def get_deployment_volume_mount_path(volume: Dict):
    if not volume.get("mount_path", None):
        return f"/mnt/{volume['type']}/{volume['name']}"
    return volume.get("mount_path")


def get_volume_cloud_resource(volume_name, volume_type):
    if volume_type == "efs":
        return get_efs_filesystem_by_name(volume_name)

    return None


def check_existing_volume_pv_and_pvc(volume_name, volume_type, env):
    if volume_type == "efs":
        return check_existing_efs_volume_k8s_resources(volume_name, env)
    return False

def create_volume_k8s_resources(volume_name, volume_type, cloud_volume_resource, env) -> bool:
    if volume_type == "efs":
        return create_efs_volume_k8s_resources(volume_name=volume_name, efs_id=cloud_volume_resource['FileSystemId'], env=env)
    else:
        raise Exception(f"Volume type '{volume_type}' is not supported")



import base64
import json
import os
from typing import Optional

import click
import requests

from tensorkube.constants import NAMESPACE, SERVICE_ACCOUNT_NAME, REGION, get_base_login_url
from tensorkube.services.eks_service import get_cluster_oidc_issuer_url
from tensorkube.services.iam_service import create_s3_csi_driver_role, attach_role_policy


def create_mountpoint_driver_role_with_policy(cluster_name, account_no, role_name, policy_name,
                                              service_account_name=SERVICE_ACCOUNT_NAME, namespace=NAMESPACE,
                                              region=REGION):
    oidc_issuer_url = get_cluster_oidc_issuer_url(cluster_name)
    create_s3_csi_driver_role(account_no, role_name, oidc_issuer_url, namespace, service_account_name)
    attach_role_policy(account_no, policy_name, role_name, region)


def get_base64_encoded_docker_config(username: str, password: str, email: str):
    auth = base64.b64encode(f"{username}:{password}".encode("utf-8")).decode("utf-8")

    docker_config_dict = {"auths": {
        "https://index.docker.io/v1/": {"username": username, "password": password, "email": email, "auth": auth, }}}

    base64_encoded_docker_config = base64.b64encode(json.dumps(docker_config_dict).encode("utf-8")).decode("utf-8")
    return base64_encoded_docker_config


def sanitise_name(name: str):
    return name.replace("_", "-").replace(" ", "-").lower()


def sanitise_assumed_role_arn(arn: str):
    arn = arn.replace('assumed-role', 'role')
    last_slash_index = arn.rfind('/')
    if last_slash_index != -1:
        arn = arn[:last_slash_index]
    return arn


def track_event(event_name: str, event_properties: dict) -> bool:
    try:
        try:
            user_email = verify_user()
        except Exception as e:
            user_email = None
        if not user_email:
            return False

        base_url = get_base_login_url()
        url = base_url + '/tensorfuse/track/tensorkube-event/'

        body = {"id": user_email, "event": event_name, "properties": event_properties}
        x = requests.post(url, json=body)
        return True
    except Exception as e:
        click.echo(f"Error while tracking event: {e}")


def extract_workdir_from_dockerfile(dockerfile_path: str) -> Optional[str]:
    with open(dockerfile_path, "r") as f:
        lines = f.readlines()
        for line in lines:
            if line.startswith("WORKDIR"):
                return line.split(" ")[1].replace("\n", "")

    return None


def extract_command_from_dockerfile(dockerfile_path: str) -> Optional[str]:
    command = None
    with open(dockerfile_path, "r") as f:
        lines = f.readlines()
        for line in lines:
            if line.startswith("CMD") or line.startswith("ENTRYPOINT"):
                command = line.split(" ")[1:]

    if not command:
        return None

    command = " ".join(command)
    command = command.replace('[', "").replace(']', "").replace('"', '').replace(",", " ").replace("\n", "")
    return command


def verify_user(session_id: str = None, token: str = None, ) -> Optional[str]:
    if not (token and session_id):
        token = os.getenv("TENSORKUBE_TOKEN", None)
        session_id = os.getenv("TENSORKUBE_SESSION_ID", None)
        if not (token and session_id):
            token_path = os.path.expanduser('~/.tensorkube/token')
            os.makedirs(os.path.dirname(token_path), exist_ok=True)

            if not os.path.exists(token_path):
                return None
            with open(token_path, "r") as f:
                token = f.readline().split("=")[1].replace("\n", "")
                session_id = f.readline().split("=")[1].replace("\n", "")

    if not (token and session_id):
        click.echo("Can't find tensorkube credentials.")
        return None

    base_url = get_base_login_url()
    response = requests.get(base_url + '/tensorfuse/verify-tensorkube-user/',
                            headers={'Content-Type': 'application/json'},
                            data=json.dumps({'session_id': session_id, 'token': token}))
    if response.status_code == 200:
        body = response.json()
        return body['email']
    else:
        return None


def save_user_credentials(session_id: str, token: str):
    token_path = os.path.expanduser('~/.tensorkube/token')
    os.makedirs(os.path.dirname(token_path), exist_ok=True)
    with open(token_path, "w") as f:
        f.write("token=" + token + "\n")
        f.write("session_id=" + session_id + "\n")



import json
import boto3
import click

from botocore.exceptions import ClientError
from pkg_resources import resource_filename
from typing import Dict

from tensorkube.constants import REGION, get_cluster_name, CliColors
from tensorkube.services.aws_service import get_aws_account_id


def create_iam_policy(policy_name: str, policy_document: Dict, region: str = REGION):
    iam = boto3.client('iam', region_name=region)
    sts = boto3.client('sts')
    try:
        # Check if the IAM policy already exists
        account_id = sts.get_caller_identity()['Account']
        policy_arn = f"arn:aws:iam::{account_id}:policy/{policy_name}"
        response = iam.get_policy(PolicyArn=policy_arn)
        print(f"IAM policy {policy_name} already exists. Skipping creation.")
        return response
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            print(f"IAM policy {policy_name} does not exist. Proceeding with creation.")
            response = iam.create_policy(PolicyName=policy_name, PolicyDocument=json.dumps(policy_document), )
            print(f"IAM policy {policy_name} created successfully.")
            return response
        else:
            print(f"An error occurred: {e}")
            raise e


def create_mountpoint_iam_policy(policy_name, bucket_name, region=REGION):
    policy_file_path = resource_filename('tensorkube', 'configurations/aws_configs/mountpoint_policy.json')
    with open(policy_file_path, 'r') as f:
        policy = json.load(f)
    for statement in policy['Statement']:
        statement['Resource'] = [r.replace('USER_BUCKET', bucket_name) for r in statement['Resource']]

    return create_iam_policy(policy_name, policy, region)


def create_sqs_access_policy(policy_name, region=REGION):
    policy_file_path = resource_filename('tensorkube', 'configurations/aws_configs/sqs_access_policy.json')
    with open(policy_file_path, 'r') as f:
        policy = json.load(f)
    for statement in policy['Statement']:
        statement['Resource'] = f"arn:aws:sqs:{region}:{get_aws_account_id()}:{get_cluster_name()}-*"
    return create_iam_policy(policy_name, policy, region)

def create_dynamo_access_policy(policy_name, region=REGION):
    policy_file_path = resource_filename('tensorkube', 'configurations/aws_configs/dynamo_access_policy.json')
    with open(policy_file_path, 'r') as f:
        policy = json.load(f)
    for statement in policy['Statement']:
        resource = statement['Resource']
        resource = resource.replace('REGION', region)
        resource = resource.replace('ACCOUNT_NO', get_aws_account_id())
        resource = resource.replace('TABLE_NAME', f"{get_cluster_name()}-*")
        statement['Resource'] = resource
    return create_iam_policy(policy_name, policy, region)



def create_iam_role_with_service_account_cluster_access(account_no: str, oidc_issuer_url: str, role_name: str,
                                                        service_account_name: str, namespace: str, region: str=REGION):
    oidc_issuer = oidc_issuer_url[8:]
    region = oidc_issuer.split('.')[2]
    trust_policy_file_path = resource_filename('tensorkube',
                                               'configurations/aws_configs/iam_role_cluster_access_trust_policy.json')
    with open(trust_policy_file_path, 'r') as f:
        trust_policy = json.load(f)
    trust_policy['Statement'][0]['Principal']['Federated'] = 'arn:aws:iam::{}:oidc-provider/{}'.format(account_no,
                                                                                                       oidc_issuer)
    trust_policy['Statement'][0]['Condition']['StringEquals'] = {
        "{}:sub".format(oidc_issuer): "system:serviceaccount:{}:{}".format(namespace, service_account_name),
        "{}:aud".format(oidc_issuer): "sts.amazonaws.com"}

    iam = boto3.client('iam', region_name=region)

    try:
        # Check if the IAM role already exists
        response = iam.get_role(RoleName=role_name)
        print(f"IAM role {role_name} already exists. Skipping creation.")
        return response
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            print(f"IAM role {role_name} does not exist. Proceeding with creation.")
            response = iam.create_role(RoleName=role_name, AssumeRolePolicyDocument=json.dumps(trust_policy), )
            print(f"IAM role {role_name} created successfully.")
            return response
        else:
            print(f"An error occurred: {e}")
            raise e



def create_or_update_iam_role_with_service_account_cluster_access(account_no: str, oidc_issuer_url: str, role_name: str,
                                                        service_account_name: str, namespace: str, region: str=REGION):
    oidc_issuer = oidc_issuer_url[8:]
    region = oidc_issuer.split('.')[2]
    trust_policy_file_path = resource_filename('tensorkube',
                                               'configurations/aws_configs/iam_role_cluster_access_trust_policy.json')
    with open(trust_policy_file_path, 'r') as f:
        trust_policy = json.load(f)
    trust_policy['Statement'][0]['Principal']['Federated'] = 'arn:aws:iam::{}:oidc-provider/{}'.format(account_no,
                                                                                                       oidc_issuer)
    trust_policy['Statement'][0]['Condition']['StringEquals'] = {
        "{}:sub".format(oidc_issuer): "system:serviceaccount:{}:{}".format(namespace, service_account_name),
        "{}:aud".format(oidc_issuer): "sts.amazonaws.com"
    }

    iam = boto3.client('iam', region_name=region)

    try:
        # Check if the IAM role already exists
        response = iam.get_role(RoleName=role_name)
        print(f"IAM role {role_name} already exists. Updating trust policy.")

        # Get the current trust policy
        current_trust_policy = response['Role']['AssumeRolePolicyDocument']
        current_sub_values = current_trust_policy['Statement'][0]['Condition']['StringEquals'].get("{}:sub".format(oidc_issuer), [])
        # If it's a string, convert it into a list for consistency
        if isinstance(current_sub_values, str):
            current_sub_values = [current_sub_values]

        new_service_account = "system:serviceaccount:{}:{}".format(namespace, service_account_name)

       # If the service account is not already present in the list, append it
        if new_service_account not in current_sub_values:
            current_sub_values.append(new_service_account)
            trust_policy['Statement'][0]['Condition']['StringEquals']["{}:sub".format(oidc_issuer)] = current_sub_values

            # Update the trust policy with the new service account
            iam.update_assume_role_policy(RoleName=role_name, PolicyDocument=json.dumps(trust_policy))
            print(f"IAM role {role_name} trust policy updated successfully.")
        else:
            print(f"Service account {service_account_name} is already included in the trust policy. No update needed.")
        return response

    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            print(f"IAM role {role_name} does not exist. Proceeding with creation.")
            response = iam.create_role(RoleName=role_name, AssumeRolePolicyDocument=json.dumps(trust_policy), )
            print(f"IAM role {role_name} created successfully.")
            return response
        else:
            print(f"An error occurred: {e}")
            raise e


def create_s3_csi_driver_role(account_no: str, role_name: str, oidc_issuer_url: str, namespace: str,
                              service_account_name: str):
    create_iam_role_with_service_account_cluster_access(account_no, oidc_issuer_url, role_name, service_account_name,
                                                        namespace)


def create_sqs_access_role(account_no: str, oidc_issuer_url: str, role_name: str, namespace: str,
                           service_account_name: str):
    return create_iam_role_with_service_account_cluster_access(account_no, oidc_issuer_url, role_name, service_account_name,
                                                        namespace)


def attach_role_policy(account_no, policy_name, role_name, region=REGION):
    client = boto3.client('iam', region_name=region)
    response = client.attach_role_policy(PolicyArn='arn:aws:iam::{}:policy/{}'.format(account_no, policy_name),
                                         RoleName=role_name, )
    return response


def detach_role_policy(account_no, role_name, policy_name, region=REGION):
    client = boto3.client('iam', region_name=region)
    response = client.detach_role_policy(PolicyArn='arn:aws:iam::{}:policy/{}'.format(account_no, policy_name),
                                         RoleName=role_name, )
    return response


def delete_role(role_name, region=REGION):
    client = boto3.client('iam', region_name=region)
    response = client.delete_role(RoleName=role_name)
    return response


def delete_policy(account_no, policy_name, region=REGION):
    client = boto3.client('iam', region_name=region)
    policy_arn = f'arn:aws:iam::{account_no}:policy/{policy_name}'
    #delete all policy versions
    policy = client.get_policy(PolicyArn=policy_arn)
    default_version = policy['Policy']['DefaultVersionId']
    # retrieve all policy versions
    versions = client.list_policy_versions(PolicyArn=policy_arn)
    for version in versions['Versions']:
        if version['VersionId'] != default_version:
            client.delete_policy_version(PolicyArn=policy_arn, VersionId=version['VersionId'])
    response = client.delete_policy(PolicyArn=policy_arn)
    return response


def update_iam_policy(account_no, policy_name, policy_document, region=REGION):
    client = boto3.client('iam', region_name=region)
    response = client.create_policy_version(PolicyArn='arn:aws:iam::{}:policy/{}'.format(account_no, policy_name),
                                           PolicyDocument=json.dumps(policy_document), SetAsDefault=True)
    return response


def delete_iam_role(role_name):
    iam_client = boto3.client('iam')
    attached_policies = iam_client.list_attached_role_policies(RoleName=role_name)['AttachedPolicies']

    # Detach policies
    for policy in attached_policies:
        iam_client.detach_role_policy(RoleName=role_name, PolicyArn=policy['PolicyArn'])
        print(f"Detached policy {policy['PolicyArn']} from role {role_name}")

    # Delete the role
    iam_client.delete_role(RoleName=role_name)
    print(f"IAM Role {role_name} deleted")


def get_role_name_for_prefix(prefix: str):
    iam = boto3.client('iam')
    roles = []
    response = iam.list_roles()
    
    # Collect all roles, handling pagination
    while True:
        roles.extend(response['Roles'])
        if response.get('IsTruncated'):  # If there are more roles
            response = iam.list_roles(Marker=response['Marker'])
        else:
            break
    
    for role in roles:
        if role['RoleName'].startswith(prefix):
            return role['RoleName']
    return None

def create_iam_user(user_name, region=REGION):
    iam = boto3.client('iam', region_name=region)
    try:
        response = iam.create_user(UserName=user_name)
        print(f"IAM User '{user_name}' created successfully.")
        return response['User']
    except iam.exceptions.EntityAlreadyExistsException:
        print(f"IAM User '{user_name}' already exists.")
        return iam.get_user(UserName=user_name)['User']

def create_eks_access_policy(policy_name, region=REGION):
    policy_file_path = resource_filename('tensorkube', 'configurations/aws_configs/eks_access_policy.json')
    with open(policy_file_path, 'r') as f:
        policy = json.load(f)
    statements= []
    for statement in policy['Statement']:
        resource = statement['Resource'] 
        resource = resource.replace('CLUSTER_NAME', get_cluster_name())
        resource =resource.replace('REGION', region)
        resource = resource.replace('ACCOUNT_NO', get_aws_account_id())
        statement['Resource'] = resource
        statements.append(statement)
    policy['Statement'] = statements
    return create_iam_policy(policy_name, policy, region)


def create_user_sqs_access_policy(policy_name, region=REGION):
    policy_file_path = resource_filename('tensorkube', 'configurations/aws_configs/user_sqs_access_policy.json')
    with open(policy_file_path, 'r') as f:
        policy = json.load(f)
    for statement in policy['Statement']:
        statement['Resource'] = f"arn:aws:sqs:{region}:{get_aws_account_id()}:{get_cluster_name()}-*"
    return create_iam_policy(policy_name, policy, region)

def create_user_s3_access_policy(policy_name, region=REGION):
    policy_file_path = resource_filename('tensorkube', 'configurations/aws_configs/user_s3_access_policy.json')
    with open(policy_file_path, 'r') as f:
        policy = json.load(f)
    return create_iam_policy(policy_name, policy, region)

def create_user_dynamo_access_policy(policy_name, region=REGION):
    policy_file_path = resource_filename('tensorkube', 'configurations/aws_configs/user_dynamo_access_policy.json')
    with open(policy_file_path, 'r') as f:
        policy = json.load(f)
    for statement in policy['Statement']:
        statement['Resource'] = f"arn:aws:dynamodb:{region}:{get_aws_account_id()}:table/{get_cluster_name()}-*"
    return create_iam_policy(policy_name, policy, region)



def attach_policy_to_user(policy_arn, user_name, region=REGION):
    iam = boto3.client('iam', region_name=region)
    iam.attach_user_policy(UserName=user_name, PolicyArn=policy_arn)
    print(f"Policy '{policy_arn}' attached to user '{user_name}'.")


def create_access_key(user_name, region=REGION):
    iam = boto3.client('iam', region_name=region)
    response = iam.create_access_key(UserName=user_name)
    access_key = response['AccessKey']
    return access_key['AccessKeyId'], access_key['SecretAccessKey']


def get_policy_source_arn():
    sts = boto3.client('sts')
    caller_arn = sts.get_caller_identity()['Arn']
    if caller_arn.endswith(':root'):
        return caller_arn
    elif ":user/" in caller_arn:
        return caller_arn
    elif ":assumed-role/" in caller_arn:
        role_name = caller_arn.split('/')[1]
        account_id = caller_arn.split(':')[4]
        return f"arn:aws:iam::{account_id}:role/{role_name}"
    else:
        return None


def has_simulate_principal_policy_permissions(policy_source_arn: str) -> bool:
    iam_client = boto3.client('iam')
    try:
        test_action = ['iam:SimulatePrincipalPolicy']
        test_resource = ['*']
        iam_client.simulate_principal_policy(
            PolicySourceArn=policy_source_arn,
            ActionNames=test_action,
            ResourceArns=test_resource
        )
        return True
    except ClientError as e:
        if e.response['Error']['Code'] == 'AccessDenied':
            click.echo(click.style("You do NOT have permissions to call 'iam:SimulatePrincipalPolicy' on Resource '*'.", fg=CliColors.ERROR.value))
            click.echo(click.style("Tensorkube uses the 'iam:SimulatePrincipalPolicy' to check for all other permissions required for running the configure command. You can refer to our docs for the full list of permissions we check for, here: https://tensorfuse.io/docs/reference/permissions/configure_permissions", fg=CliColors.WARNING.value))
            click.echo(click.style("Please request the necessary IAM permissions before running this check.", fg=CliColors.ERROR.value))
            return False
        else:
            raise


def has_configure_permissions() -> bool:
    click.echo("Checking if you have the necessary AWS permissions to run the configure command...")

    policy_source_arn = get_policy_source_arn()
    if not policy_source_arn:
        raise Exception("Could not determine the policy source ARN.")

    elif  policy_source_arn.endswith(':root'):
        click.echo(click.style("You are running this as the root user.", fg=CliColors.WARNING.value))
        click.echo(click.style("Skipping 'SimulatePrincipalPolicy' check for root ARN as it is not possible to run this check for the root user.", fg=CliColors.WARNING.value))
        click.echo(click.style("Please ensure no AWS Organizations Service Control Policies limit your root account.", fg=CliColors.WARNING.value))
        click.echo(click.style("You can refer to our docs for the full list of permissions we check for, here: https://tensorfuse.io/docs/reference/permissions/configure_permissions", fg=CliColors.WARNING.value))
        return True


    click.echo(f"Running tests for: {policy_source_arn}")
    if not has_simulate_principal_policy_permissions(policy_source_arn):
        return False


    policy_file_path = resource_filename('tensorkube', 'permissions/configure_permissions.json')
    with open(policy_file_path, 'r') as f:
        policy_doc = json.load(f)

    statements = policy_doc.get("Statement", [])
    if not isinstance(statements, list):
        # Some policies have Statement as dict if there's only one.
        statements = [statements]

    iam_client = boto3.client('iam')
    account_no = get_aws_account_id()
    all_allowed = True
    for idx, statement in enumerate(statements):
        actions = statement.get("Action", [])
        resources = statement.get("Resource", [])

        if isinstance(actions, str):
            actions = [actions]
        if isinstance(resources, str):
            resources = [resources]
        resources = [resource.replace('<ACCOUNT_NO>', account_no) for resource in resources]

        try:
            response = iam_client.simulate_principal_policy(
                PolicySourceArn=policy_source_arn,
                ActionNames=actions,
                ResourceArns=resources
            )
        except ClientError as e:
            print(f"Error simulating statement: {e}")
            continue

        eval_results = response.get('EvaluationResults', [])
        for result in eval_results:
            action_name = result.get('EvalActionName')
            resource_name = result.get('EvalResourceName')
            decision = result.get('EvalDecision')
            if decision != 'allowed':
                click.echo(click.style(f"Permissions Missing. Action: {action_name}, Resource: {resource_name} => Decision: {decision}", fg=CliColors.ERROR.value))
                all_allowed = False

    return all_allowed



import importlib.metadata
from importlib import resources
from typing import Optional

import click
from kubernetes import config, client
from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name
from tensorkube.services.version_service import get_cluster_tensorkube_version, get_current_tensorkube_version



def version_to_tuple(version_str):
    # Split the version string by '.' and convert each component to an integer
    return tuple(map(int, version_str.split('.')))


def set_current_cli_version_to_cluster(context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    config_map = client.V1ConfigMap(metadata=client.V1ObjectMeta(name="tensorkube-migration", namespace="default"),
                                    data={"version": get_current_tensorkube_version()})
    try:
        v1.replace_namespaced_config_map("tensorkube-migration", "default", config_map)
    except client.exceptions.ApiException as e:
        if e.status == 404:
            v1.create_namespaced_config_map("default", config_map)
        else:
            raise e





def load_migrations():
    # Get a reference to the migrations directory
    migrations_dir = resources.files('tensorkube.migration_service').joinpath('migrations')

    # List all Python files in the directory, excluding __init__.py
    migration_files = [f.name for f in migrations_dir.iterdir() if f.name.endswith('.py') and f.name != '__init__.py']

    # Sort migrations based on the leading numbers in the filenames
    migrations = sorted(migration_files, key=lambda x: tuple(map(int, x.split('_')[:3])))

    cluster_tensorkube_version = get_cluster_tensorkube_version()
    click.echo(f"Cluster TensorKube version: {cluster_tensorkube_version}")
    if cluster_tensorkube_version is None:
        cluster_tensorkube_version = '0.0.0'

    migrations = filter(lambda migration: tuple(map(int, migration.split('_')[:3])) > version_to_tuple(cluster_tensorkube_version),
        migrations)

    return [migration.replace('.py', '') for migration in migrations]


def apply_migration(migration_name):
    # Adjust the module path to reflect the structure starting from the project root
    migration_module_path = f"tensorkube.migration_service.migrations.{migration_name}"
    migration_module = importlib.import_module(migration_module_path)
    migration_module.apply()


def migrate_tensorkube():
    migrations = load_migrations()
    for migration in migrations:
        click.echo(f"Applying migration {migration}...")
        apply_migration(migration)
    set_current_cli_version_to_cluster()
    click.echo("Migration completed successfully.")
    cluster_tensorkube_version = get_cluster_tensorkube_version()
    click.echo(f"Cluster TensorKube version: {cluster_tensorkube_version}")

def test_mig():
    migrations = load_migrations()[0]
    migration_module_path = f"tensorkube.migration_service.migrations.{migrations}"
    migration_module = importlib.import_module(migration_module_path)
    migration_module.test_migration()



import boto3
import os
import yaml
import cfnresponse
import subprocess
from kubernetes import client, config
from typing import Dict
from enum import Enum
from time import sleep, time


KUBECONFIG_FILE_PATH = '/tmp/kubeconfig'

class Operation(Enum):
    UPDATE_AUTHMAP = 'update-authmap'
    HELMFILE_SYNC = 'helmfile-sync'
    PATCH_SERVICE_ACCOUNT = 'patch-service-account'
    SYNC_CLI_VERSION_TO_CLUSTER = 'sync-cli-version-to-cluster'
    CREATE_ENV = 'create-env'
    TEARDOWN_KSVCS_SCALEDJOBS = 'teardown-ksvcs-scaledjobs'


def get_aws_caller_identity():
    return boto3.client('sts').get_caller_identity()


def generate_kubeconfig(cluster_name: str, region: str, aws_account_id: str):
    eks_client = boto3.client('eks', region_name=region)

    cluster_info = eks_client.describe_cluster(name=cluster_name)['cluster']

    # Extract cluster details
    endpoint = cluster_info['endpoint']
    cluster_ca_certificate = cluster_info['certificateAuthority']['data']

    # Create a kubeconfig dictionary
    kubeconfig_data = {
        "apiVersion": "v1",
        "clusters": [{
            "cluster": {
                "server": endpoint,
                "certificate-authority-data": cluster_ca_certificate
            },
            "name": f'arn:aws:eks:us-east-1:{aws_account_id}:cluster/{cluster_name}'
        }],
        "contexts": [{
            "context": {
                "cluster": f'arn:aws:eks:us-east-1:{aws_account_id}:cluster/{cluster_name}',
                "user": f'arn:aws:eks:us-east-1:{aws_account_id}:cluster/{cluster_name}'
            },
            "name": f'arn:aws:eks:us-east-1:{aws_account_id}:cluster/{cluster_name}'
        }],
        "current-context": f'arn:aws:eks:us-east-1:{aws_account_id}:cluster/{cluster_name}',
        "kind": "Config",
        "users": [{
            "name": f'arn:aws:eks:us-east-1:{aws_account_id}:cluster/{cluster_name}',
            "user": {
                "exec": {
                    "apiVersion": "client.authentication.k8s.io/v1beta1",
                    "command": "aws",
                    "args": [
                        "eks",
                        "get-token",
                        "--cluster-name", cluster_name
                    ],
                    "env": [
                        {"name": "AWS_DEFAULT_REGION", "value": region}
                    ]
                }
            }
        }]
    }
    # kubeconfig_dir = os.path.expanduser('~/.kube')
    # kubeconfig_file = os.path.join(kubeconfig_dir, 'config')
    kubeconfig_dir = os.path.expanduser('/tmp')
    kubeconfig_file = os.path.join(kubeconfig_dir, 'kubeconfig')

    # Create the directory if it does not exist
    if not os.path.exists(kubeconfig_dir):
        os.makedirs(kubeconfig_dir, exist_ok=True)

    # Load existing kubeconfig or create a new one if it doesn't exist
    if os.path.exists(kubeconfig_file):
        with open(kubeconfig_file, 'r') as f:
            existing_config = yaml.safe_load(f) or {}
    else:
        existing_config = {'apiVersion': 'v1', 'kind': 'Config', 'clusters': [], 'contexts': [], 'users': []}

    # Merge new kubeconfig entry into existing config
    # Function to update or add an entry in the list
    def update_or_add_entry(existing_list, new_entry, key):
        for i, entry in enumerate(existing_list):
            if entry[key] == new_entry[key]:
                existing_list[i] = new_entry  # Update the existing entry
                return
        existing_list.append(new_entry)  # Add as a new entry if not found

    # Merge new kubeconfig entry into existing config
    update_or_add_entry(existing_config['clusters'], kubeconfig_data['clusters'][0], 'name')
    update_or_add_entry(existing_config['contexts'], kubeconfig_data['contexts'][0], 'name')
    update_or_add_entry(existing_config['users'], kubeconfig_data['users'][0], 'name')
    existing_config['current-context'] = kubeconfig_data['current-context']

    # Write back to the default kubeconfig file
    with open(kubeconfig_file, 'w') as f:
        yaml.dump(existing_config, f)

    os.chmod(kubeconfig_file, 0o700)


def get_cluster_info(cluster_name: str):
    eks = boto3.client('eks')
    cluster = eks.describe_cluster(name=cluster_name)['cluster']
    return {
        'endpoint': cluster['endpoint'],
        'ca': cluster['certificateAuthority']['data']
    }


def add_karpenter_role_to_aws_auth_config_map(role_arn: str):
    # Load the kubeconfig
    # kubeconfig_file = os.path.expanduser('~/.kube/config')
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    # Load the existing aws-auth config map
    v1 = client.CoreV1Api()
    aws_auth_config_map = v1.read_namespaced_config_map(name='aws-auth', namespace='kube-system')

    # Extract the existing roles
    roles = yaml.safe_load(aws_auth_config_map.data['mapRoles'])

    # Add the new role
    roles.append({
        "rolearn": role_arn,
        "username": "system:node:{{EC2PrivateDNSName}}",
        "groups": ["system:bootstrappers", "system:nodes"]
    })

    # Update the aws-auth config map
    aws_auth_config_map.data['mapRoles'] = yaml.dump(roles)
    v1.patch_namespaced_config_map(name='aws-auth', namespace='kube-system', body=aws_auth_config_map)


def update_values_yaml(parameters: Dict):
    cluster_name = parameters.get('ClusterName')
    sqs_access_role_arn = parameters.get('SQSAccessRoleArn')
    region = parameters.get('Region')
    bucket_name = parameters.get('DefaultEnvBuildBucketName')
    buildkit_isra_role_arn = parameters.get('BuildkitISRARoleArn')
    job_queue_sidecar_role_arn = parameters.get('JobQueueSidecarIAMRoleArn')


    if not (cluster_name and sqs_access_role_arn and region and bucket_name and buildkit_isra_role_arn and job_queue_sidecar_role_arn):
        raise Exception("ClusterName, SQSAccessRoleArn, Region, DefaultEnvBuildBucketName, JobQueueSidecarIAMRoleArn and BuildkitISRARoleArn are required parameters")

    with open("helm-charts/values.yaml", "r") as f:
        values = yaml.safe_load(f) or {}

    values['clusterName'] = cluster_name
    values['sqsAccessRoleArn'] = sqs_access_role_arn
    values['region'] = region
    values['envName'] = "default"
    values['defaultEnvBuildBucketName'] = bucket_name
    values['buildkitISRARoleArn'] = buildkit_isra_role_arn
    values['jobQueueISRARoleArn'] = job_queue_sidecar_role_arn
    values['managedNodeGroupName'] = f"{cluster_name}-ng-bottlerocket"

    with open("/tmp/values.yaml", "w") as f:
        yaml.dump(values, f)


def run_helmfile_sync(parameters: Dict):
    update_values_yaml(parameters=parameters)
    subprocess.run(["helmfile", "sync", "--file", "./helm-charts/helmfile.yaml"], check=True)


def run_helmfile_destroy(parameters: Dict):
    update_values_yaml(parameters=parameters)
    tries = 0
    while tries < 3:
        try:
            subprocess.run(["helmfile", "destroy", "--file", "./helm-charts/helmfile.yaml"], check=True)
            return True
        except Exception as e:
            tries += 1
            print(f"An error occurred: {e}")
            print("Retrying in 30 seconds...")
            sleep(30)

    print("Failed to destroy helm resources after 3 attempts.")
    return False


def patch_service_account(role_arn: str, service_account_name: str, namespace: str):
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    # Load the existing service account
    v1 = client.CoreV1Api()
    service_account = v1.read_namespaced_service_account(name=service_account_name, namespace=namespace)

    # Extract the existing annotations
    annotations = service_account.metadata.annotations or {}

    # Add the new annotation
    annotations['eks.amazonaws.com/role-arn'] = role_arn

    # Update the service account
    service_account.metadata.annotations = annotations
    v1.patch_namespaced_service_account(name=service_account_name, namespace=namespace, body=service_account)


def set_current_cli_version_to_cluster(current_version: str):
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    v1 = client.CoreV1Api()
    config_map = client.V1ConfigMap(
        metadata=client.V1ObjectMeta(name="tensorkube-migration", namespace="default"),
        data={"version": current_version}
    )
    try:
        v1.replace_namespaced_config_map("tensorkube-migration", "default", config_map)
    except client.exceptions.ApiException as e:
        if e.status == 404:
            v1.create_namespaced_config_map("default", config_map)
        else:
            raise e


def create_env(env_name: str, build_bucket_name: str, buildkit_isra_role_arn: str, region: str, cluster_name: str):
    try:
        subprocess.run(["helm", "install", f"{env_name}-env", "./helm-charts/tk-env", "--set", f"envName={env_name}",
                        "--set", f"bucketName={build_bucket_name}", "--set", f"region={region}",
                        "--set", f"clusterName={cluster_name}", "--set", f"buildkitISRARoleArn={buildkit_isra_role_arn}"
                        ], check=True)
        return True
    except Exception as e:
        print(f"An error occurred: {e}")
        return False


def delete_all_ksvcs() -> bool:
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    # Create the CustomObjectsApi
    custom_api = client.CustomObjectsApi()
    core_v1_api = client.CoreV1Api()

    # Group, version, and plural for Knative Services CRD
    group = "serving.knative.dev"
    version = "v1"
    plural = "services"

    namespaces = core_v1_api.list_namespace()
    for ns in namespaces.items:
        ns_name = ns.metadata.name

        try:
            # List all Knative Services in the current namespace
            knative_services = custom_api.list_namespaced_custom_object(
                group=group,
                version=version,
                namespace=ns_name,
                plural=plural
            )

            # Delete each Knative Service found
            for svc in knative_services.get("items", []):
                svc_name = svc["metadata"]["name"]
                print(f"Deleting Knative Service '{svc_name}' in namespace '{ns_name}'...")
                custom_api.delete_namespaced_custom_object(
                    group=group,
                    version=version,
                    namespace=ns_name,
                    plural=plural,
                    name=svc_name
                )

        except client.exceptions.ApiException as e:
            print(f"Error deleting ksvc in namespace '{ns_name}': {e}")
            return False
    return True


def delete_all_scaledjobs():
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    custom_api = client.CustomObjectsApi()
    group = "keda.sh"
    version = "v1alpha1"
    plural = "scaledjobs"
    namespace = "keda"

    try:
        scaled_jobs = custom_api.list_namespaced_custom_object(
            group=group,
            version=version,
            namespace=namespace,
            plural=plural
        )

        # Delete each ScaledJob found
        for sj in scaled_jobs.get("items", []):
            sj_name = sj["metadata"]["name"]
            print(f"Deleting ScaledJob '{sj_name}' in namespace '{namespace}'...")
            custom_api.delete_namespaced_custom_object(
                group=group,
                version=version,
                namespace=namespace,
                plural=plural,
                name=sj_name
            )

    except client.exceptions.ApiException as e:
       return False
    return True


def wait_for_nodes_with_label_only(cluster_name: str, timeout_seconds: int = 840,
                                   poll_interval_seconds: int = 10) -> True:
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    v1 = client.CoreV1Api()

    # Label key and value we want to keep
    label_key = "alpha.eksctl.io/nodegroup-name"
    label_value = f"{cluster_name}-ng-bottlerocket"

    start_time = time()

    while True:
        nodes = v1.list_node().items

        remaining_unlabeled = [
            node for node in nodes
            if node.metadata.labels.get(label_key, None) != label_value
        ]

        if not remaining_unlabeled:
            print("All nodes without the desired label have been removed.")
            return True

        elapsed = time() - start_time
        if elapsed > timeout_seconds:
            raise TimeoutError(
                f"Timed out after {timeout_seconds} seconds waiting for all other nodes to be deleted."
            )

        print(
            f"Still waiting for {len(remaining_unlabeled)} node(s) without label "
            f"'{label_key}={label_value}' to be deleted. Retrying in {poll_interval_seconds}s..."
        )
        sleep(poll_interval_seconds)


def delete_all_ksvcs_scaledjobs(parameters: Dict) -> bool:
    #TODO?: delete all build jobs?
    cluster_name = parameters.get('ClusterName')
    if not cluster_name:
        raise Exception("ClusterName is a required parameter")

    deleted_ksvcs = delete_all_ksvcs()
    if not deleted_ksvcs:
        return False

    deleted_scaledjobs = delete_all_scaledjobs()
    if not deleted_scaledjobs:
        return False

    try:
        wait_for_nodes_with_label_only(cluster_name=cluster_name)
    except Exception as e:
        print(f"An error occurred: {e}")
        return False

    return True


def perform_create_operation(operation: str, parameters: Dict) -> bool:
    if operation == Operation.UPDATE_AUTHMAP.value:
        karpenter_role_arn = parameters.get('KarpenterRoleArn')
        add_karpenter_role_to_aws_auth_config_map(karpenter_role_arn)
        return True
    elif operation == Operation.HELMFILE_SYNC.value:
        run_helmfile_sync(parameters)
        return True
    elif operation == Operation.PATCH_SERVICE_ACCOUNT.value:
        role_arn = parameters.get('RoleArn')
        service_account_name = parameters.get('ServiceAccountName')
        namespace = parameters.get('Namespace')
        if not role_arn or not service_account_name or not namespace:
            raise Exception("RoleArn, ServiceAccountName and Namespace are required parameters")
        patch_service_account(role_arn, service_account_name, namespace)
        return True
    elif operation == Operation.SYNC_CLI_VERSION_TO_CLUSTER.value:
        cli_version = parameters.get('CliVersion')
        if not cli_version:
            raise Exception("CliVersion is a required parameter")
        set_current_cli_version_to_cluster(cli_version)
        return True
    elif operation == Operation.CREATE_ENV.value:
        env_name = parameters.get('EnvName')
        build_bucket_name = parameters.get('BuildBucketName')
        buildkit_isra_role_arn = parameters.get('BuildkitISRARoleArn')
        region = parameters.get('Region')
        cluster_name = parameters.get('ClusterName')
        if not (env_name and build_bucket_name and buildkit_isra_role_arn and region and cluster_name):
            raise Exception("EnvName, BuildBucketName, BuildkitISRARoleArn, Region and ClusterName are required parameters")
        return create_env(env_name, build_bucket_name, buildkit_isra_role_arn, region, cluster_name)
    elif operation in [Operation.TEARDOWN_KSVCS_SCALEDJOBS.value]:
        return True
    else:
        return False


def perform_delete_operation(operation: str, parameters: Dict) -> bool:
    if operation == Operation.HELMFILE_SYNC.value:
        return run_helmfile_destroy(parameters)
    elif operation == Operation.TEARDOWN_KSVCS_SCALEDJOBS.value:
        return delete_all_ksvcs_scaledjobs(parameters)
    elif operation in [Operation.UPDATE_AUTHMAP.value, Operation.PATCH_SERVICE_ACCOUNT.value,
                       Operation.SYNC_CLI_VERSION_TO_CLUSTER.value, Operation.CREATE_ENV.value]:
        return True
    else:
        return False


def handler(event, context):
    cluster_name = event['ResourceProperties'].get('ClusterName')
    command = event['ResourceProperties'].get('Operation')
    parameters = event['ResourceProperties'].get('Parameters')
    region = event['ResourceProperties'].get('Region')
    aws_account_id = event['ResourceProperties'].get('AWSAccountId')

    print('Generating kubeconfig')
    generate_kubeconfig(cluster_name, region, aws_account_id)
    os.environ['XDG_CONFIG_HOME'] = '/tmp/.config'
    os.environ['XDG_CACHE_HOME'] = '/tmp/.cache'
    os.environ['KUBECONFIG'] = KUBECONFIG_FILE_PATH
    print('Kubeconfig generated successfully')
    print("testing v3")

    if event['RequestType'] == 'Create':
        try:
            success = perform_create_operation(command, parameters)

            if success:
                cfnresponse.send(event, context, cfnresponse.SUCCESS, {'Event': 'Create', 'Reason': 'Operation successful'})
            else:
                cfnresponse.send(event, context, cfnresponse.FAILED, {'Event': 'Create', 'Reason': 'Invalid operation or parameters'})
        except Exception as e:
            cfnresponse.send(event, context, cfnresponse.FAILED, {'Event': 'Create', 'Reason': str(e)})
    elif event['RequestType'] == 'Update':
        cfnresponse.send(event, context, cfnresponse.SUCCESS, {'Event': 'Update', 'Reason': 'No action needed'})
    elif event['RequestType'] == 'Delete':
        try:
            success = perform_delete_operation(command, parameters)

            if success:
                cfnresponse.send(event, context, cfnresponse.SUCCESS, {'Event': 'Delete', 'Reason': 'Operation successful'})
            else:
                cfnresponse.send(event, context, cfnresponse.FAILED, {'Event': 'Delete', 'Reason': 'Invalid operation or parameters'})
        except Exception as e:
            cfnresponse.send(event, context, cfnresponse.FAILED, {'Event': 'Delete', 'Reason': str(e)})
    else:
        cfnresponse.send(event, context, cfnresponse.FAILED, {'Reason': 'Invalid request type'})



from enum import Enum
from typing import Optional

REGION = 'us-east-1'
NAMESPACE = 'kube-system'
DEFAULT_NAMESPACE = 'default'
SERVICE_ACCOUNT_NAME = 's3-csi-driver-sa'
ADDON_NAME = "aws-mountpoint-s3-csi-driver"
KNATIVE_SERVING_NAMESPACE = 'knative-serving'
CONFIG_FEATURES = 'config-features'
BUILD_TOOL = 'buildkit'
DATASET_BUCKET_TYPE = 'datasets'

DEFAULT_CAPABILITIES = ['CAPABILITY_NAMED_IAM']
TENSORKUBE_MACRO_NAME = "TensorkubePyPlate"

KEDA_NAMESPACE = 'keda'
KEDA_SERVICE_ACCOUNT_NAME = 'keda-operator'
KEDA_IAM_ROLE_NAME = 'keda-iam-role'

CREATED_BY_TAG = 'CreatedBy'
CLUSTER_NAME_TAG = 'ClusterName'
TENSORFUSE_STRING = 'tensorfuse'


TENSORFUSE_NAMESPACES = ['amazon-cloudwatch', 'kube-system', 'istio-system', 'knative-serving', 'kube-node-lease',
                         'kube-public', 'kube-system', 'nydus-snapshotter' ]

LOCKED_AWS_CLI_VERSION = "2.22.8"
LOCKED_EKSCTL_VERSION = "0.194.0"
LOCKED_KUBECTL_VERSION = "1.31.3"
LOCKED_HELM_VERSION = "3.16.3"
LOCKED_ISTIO_VERSION = "1.24.0"


NODEPOOL = 'nodepool'
EC2NODECLASS = 'ec2nodeclass'
GPU_NODEPOOL_VERSION = 'gpu-v1'

AWS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION = 'v1.0.1'
EKS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION = 'v1.0.0'

GPU_DEPLOYMENT_MIN_CPU = 2000
GPU_DEPLOYMENT_MIN_MEMORY = 8000

def get_cluster_name():
    return "tensorkube"

def get_cfn_base_stack_name():
    return f"{get_cluster_name()}-base-stack"

def get_template_bucket_name(test: bool):
    if test:
        return "tensorkube-cfn-staging-70w939rx1f"
    return "tensorkube-cfn-templates-bwgkphrtz5pdbv"


def get_image_registry_id(test):
    if test:
        return "s6z9f6e5"
    return "q0m5o9l2"


def get_templates_version():
    return "v0.0.2"


def get_mount_policy_name(cluster_name: str, env: Optional[str] = None) -> str:
    if env:
        return f'{cluster_name}-mountpoint-policy-env-{env}'
    return f'{cluster_name}-mountpoint-policy'


def get_mount_driver_role_name(cluster_name):
    return f'{cluster_name}-mountpoint-driver-role'


class Events(Enum):
    LOGGED_IN = 'logged-in'

    COMMAND_RUN = 'command-run'

    CONFIGURE_START = 'configure-start'
    CONFIGURE_END = 'configure-end'
    CONFIGURE_ERROR = 'configure-error'

    TEARDOWN_START = 'teardown-start'
    TEARDOWN_END = 'teardown-end'

    DEPLOY_START = 'deploy-start'
    DEPLOY_END = 'deploy-end'
    DEPLOY_ERROR = 'deploy-error'

    DEVCONTAINER = 'devcontainer'
    DEVCONTAINER_START = 'devcontainer-start'
    DEVCONTAINER_STOP = 'devcontainer-stop'
    DEVCONTAINER_LIST = 'devcontainer-list'
    DEVCONTAINER_DELETE = 'devcontainer-delete'
    DEVCONTAINER_ERROR = 'devcontainer-error'

    JOB_DEPLOY_START = 'job-deploy-start'
    JOB_DEPLOY_END = 'job-deploy-end'
    JOB_DEPLOY_ERROR = 'job-deploy-error'

    TEST_START = 'test-start'
    TEST_END = 'test-end'


class PodStatus(Enum):
    PENDING = 'Pending'
    RUNNING = 'Running'
    SUCCEEDED = 'Succeeded'
    FAILED = 'Failed'
    UNKNOWN = 'Unknown'


class CliColors(Enum):
    ERROR = 'red'
    SUCCESS = 'green'
    WARNING = 'yellow'
    INFO = 'blue'
    DEFAULT = 'white'

class TeardownType(Enum):
    LEGACY = 'legacy'
    CFN = 'cfn'
    UNKNOWN = 'unknown'

class ProbeType(Enum):
    LIVENESS = 'liveness'
    READINESS = 'readiness'
    STARTUP = 'startup'

def get_efs_service_account_name() -> str:
    return f'{get_cluster_name()}-efs-csi-controller-sa'


def get_efs_role_name() -> str:
    return f'AmazonEKS_EFS_CSI_DriverRole_{get_cluster_name()}'


def get_logging_service_account_name() -> str:
    return f'cloudwatch-agent'


def get_cloudwatch_role_name() -> str:
    return f'cloudwatch-agent-role_{get_cluster_name()}'


def get_cloudwatch_namespace() -> str:
    return 'amazon-cloudwatch'


def get_job_sidecar_iam_role_name() -> str:
    return f'{get_cluster_name()}-job-sidecar-iam-role'


def get_efs_security_group_name(env: Optional[str] = None) -> str:
    if env:
        return f'eks-efs-sg-{get_cluster_name()}-env-{env}'
    return f'eks-efs-sg-{get_cluster_name()}'


def get_base_login_url():
    return 'https://api.tensorfuse.io'



import json
import logging
import time
from typing import Optional, List

import boto3
import click
import yaml
from kubernetes import config, utils
from kubernetes.client.rest import ApiException
from pkg_resources import resource_filename

from tensorkube.constants import DEFAULT_NAMESPACE, get_cluster_name, DATASET_BUCKET_TYPE
from tensorkube.helpers import sanitise_name
from tensorkube.services.aws_service import get_bucket_name
from tensorkube.services.deploy import DEFAULT_CPU, DEFAULT_MEMORY
from tensorkube.services.job_queue_service import deploy_job, queue_job, is_existing_queued_job
from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name, create_configmap, generate_kubeconfig
from tensorkube.services.knative_service import get_instance_family_from_gpu_type
from tensorkube.services.s3_access_service import get_s3_service_account_name
from tensorkube.services.s3_service import create_s3_bucket

# disable botocore info logs
logging.getLogger('botocore').setLevel(logging.CRITICAL)

AXOLOTL_IMAGE_TRAIN_JOB = "gane5h/build_tool:axo"
AXOLOTL_IMAGE_SCALED_JOB = "tensorfuse/axolotl-train:v0.0.6"


def get_training_stats(job_name: str, start_time: int, namespace: str = "default"):
    log_group = f"/aws/containerinsights/{get_cluster_name()}/application"
    log_stream_prefix = job_name
    client = boto3.client('logs')
    # Define the log query pattern to search for
    log_pattern = 'loss'
    end_time = int(time.time())
    try:
        # Start the query
        response = client.start_query(logGroupName=log_group, startTime=start_time, endTime=end_time,
            queryString=f"fields @timestamp, @message | filter @message like /'{log_pattern}'/ and @logStream like /{log_stream_prefix}/ | sort @timestamp desc | limit 1")
        query_id = response["queryId"]

        while True:
            # Poll for the query results
            time.sleep(1)
            results = client.get_query_results(queryId=query_id)

            # Check if query is complete
            if results["status"] in ["Complete", "Failed", "Cancelled", "Timeout", "Unknown"]:
                break
        msg = ""
        if results["status"] == "Complete" and results["results"]:
            # Extract the relevant log from the results
            for result in results["results"]:
                message = result[1]["value"]
                message_dict = json.loads(message)
                msg = message_dict["log"]
                return msg
    except client.exceptions.ResourceNotFoundException as e:
        print(f"Resource not found: {e}")  # TODO: use good logger
    except Exception as e:
        print(f"An error occurred: {e}")  # TODO: use good logger
    return msg


def apply_training_job(job_name: str, namespace: str, yaml_file_path: str, gpus: int, gpu_type: str,
                       context_name: Optional[str] = None, secrets: List[str] = [], image_url: Optional[str] = None,
                       bucket: Optional[str] = None):
    # Load kube config
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Read the YAML file
    with open(yaml_file_path, 'r') as f:
        yaml_content = f.read()

    image_tag = image_url.split(':')[-1]
    click.echo(f"Creating job {job_name} with {gpus} GPUs.")
    yaml_content = yaml_content.replace('${JOB_NAME}', job_name)
    yaml_content = yaml_content.replace('${GPUS}', str(gpus))
    yaml_content = yaml_content.replace('${IMAGE_URL}', image_url)
    yaml_content = yaml_content.replace('${IMAGE_TAG}', image_tag)
    yaml_content = yaml_content.replace('${NAMESPACE}', namespace)
    yaml_content = yaml_content.replace('${AXOLOTL_CONFIGMAP_NAME}', f"{job_name}-config")

    # Load the YAML content    
    yaml_dict = yaml.safe_load(yaml_content)
    yaml_dict['spec']['template']['spec']['serviceAccountName'] = get_s3_service_account_name(namespace)
    if bucket:
        yaml_dict['spec']['template']['spec']['containers'][0]['env'].append(
            {'name': 'LORA_ADAPTER_BUCKET', 'value': bucket  # TODO: should we take region as well?
             })
    yaml_dict['spec']['template']['spec']['containers'][0]['env'].append({'name': 'JOB_NAME', 'value': job_name})

    if secrets:
        yaml_dict['spec']['template']['spec']['volumes'].append({'name': 'secrets', 'projected': {
            'sources': [{'secret': {'name': secret_name}} for secret_name in secrets]}})

        yaml_dict['spec']['template']['spec']['containers'][0]['volumeMounts'].append(
            {'name': 'secrets', 'mountPath': '/mnt/secrets', 'readOnly': True})

    yaml_dict['spec']['template']['spec']['nodeSelector'] = {
        'karpenter.k8s.aws/instance-family': get_instance_family_from_gpu_type(gpu_type), }

    # Create the job
    utils.create_from_dict(k8s_api_client, yaml_dict)
    click.echo(f"Job {job_name} created successfully.")


def get_job_prefix_from_job_id(job_id: str) -> str:
    job_name = sanitise_name(job_id)
    job_name = f"ax-{job_name}"
    return job_name


def get_training_id_from_job_name(job_name: str) -> str:
    return job_name.split('ax-')[1]


def validate_training_config(config: dict):
    if 'base_model' not in config:
        raise Exception("base_model is required")
    if 'datasets' not in config:
        raise Exception("datasets is required")
    return


def get_dataset_path(dataset_name: str) -> str:
    bucket_name = get_bucket_name(type=DATASET_BUCKET_TYPE)
    return f"s3://{bucket_name}/{dataset_name}.jsonl"


def get_axolotl_base_config(gpus: int) -> dict:
    fileName = resource_filename('tensorkube', 'configurations/axolotl_configs/single_gpu.yaml')
    if gpus > 1:
        fileName = resource_filename('tensorkube', 'configurations/axolotl_configs/multi_gpu.yaml')
    with open(fileName, 'r') as f:
        yaml_dict = yaml.safe_load(f)
    return yaml_dict


# todo: currently supporting one dataset inside datasets
def convert_dataset_to_s3_path(datasets: dict) -> str:
    if len(datasets) != 1:
        raise Exception("Only one dataset is supported")
    dataset = datasets[0]
    if 'id' in dataset:
        dataset['path'] = get_dataset_path(dataset['id'])
        return dataset['path']

    return dataset['path']


def axolotl_train(env: str, secrets: list, gpus: int, gpu_type: str, job_id: str, config_path: str):
    if config_path is None:
        raise Exception("config_path is required")
    if job_id is None:
        raise Exception("job_id is required")

    bucket_name = get_bucket_name(env_name=env, type='train')
    create_s3_bucket(bucket_name)
    yaml_dict = None
    with open(config_path, 'r') as f:  # user config path
        yaml_dict = yaml.safe_load(f)
    if yaml_dict is None or len(yaml_dict) == 0:
        raise Exception("Invalid yaml data")
    # validate the config
    validate_training_config(yaml_dict)
    # merge with base config
    base_yaml_dict = get_axolotl_base_config(gpus)
    yaml_dict = {**base_yaml_dict, **yaml_dict}
    # convert dataset to s3 path
    s3_path = convert_dataset_to_s3_path(yaml_dict['datasets'])
    namespace = DEFAULT_NAMESPACE if not env else env
    job_name = f'{get_job_prefix_from_job_id(job_id)}'
    try:
        config_name = f"{job_name}-config"
        create_configmap(config_name, namespace=namespace, data=yaml_dict, context_name=None, force=True)
    except Exception as e:
        raise Exception("Failed to create the configmap")

    yaml_file_path = resource_filename('tensorkube', 'configurations/build_configs/axolotl-train.yaml')

    try:
        apply_training_job(job_name, namespace, yaml_file_path, gpus, gpu_type, secrets=secrets,
                           image_url=AXOLOTL_IMAGE_TRAIN_JOB, bucket=bucket_name)
    except Exception as e:
        raise Exception("Failed to create the training job")
    return


# from hugging face model path
supported_base_models = ['meta-llama/Llama-3.1-70B-Instruct', 'meta-llama/Llama-3.1-8B-Instruct']


def get_dataset_config_for_base_model(base_model: str, s3_path: str) -> dict:
    if base_model not in supported_base_models:
        raise Exception(f"Invalid base model {base_model}")
    base_config = get_axolotl_base_config_by_model(base_model)
    base_config['datasets'][0]['path'] = s3_path
    return base_config['datasets']


def get_config_from_paramaters(base_model: str, dataset: str, epochs: int, micro_batch_size: int, learning_rate: float,
                               lora_r: int,base_config: dict,wandb_entity=Optional[str], wandb_project=Optional[str],
                                        wandb_name=Optional[str], val_set_size=Optional[float], hf_org_id: Optional[str] = None) -> dict:
    if micro_batch_size is None:
        micro_batch_size = base_config['micro_batch_size']
    if learning_rate is None:
        learning_rate = base_config['learning_rate']
    if lora_r is None:
        lora_r = base_config['lora_r']
    datasets_config = []
    if "s3://" in dataset:
        datasets_config = [{"path": dataset}]
    else:
        datasets_config = [{"id": dataset}]
    if val_set_size is None:
        val_set_size = base_config['val_set_size']
        evals_per_epoch = base_config['evals_per_epoch']
    else:
        evals_per_epoch = 1
    if wandb_entity is None or wandb_project is None:
        wandb_entity = base_config['wandb_entity']
        wandb_project = base_config['wandb_project']
    config = {"base_model": base_model, "datasets": datasets_config, "num_epochs": epochs,
        "micro_batch_size": micro_batch_size, "learning_rate": learning_rate, "lora_r": lora_r,
              "evals_per_epoch": evals_per_epoch,
              "val_set_size": val_set_size,
              "wandb_entity": wandb_entity, "wandb_project": wandb_project, "wandb_name": wandb_name, "hf_org_id": hf_org_id}
    s3_path = convert_dataset_to_s3_path(config['datasets'])
    config['datasets'] = get_dataset_config_for_base_model(base_model, s3_path)
    return config


def get_axolotl_base_config_by_model(base_model: str) -> dict:
    if base_model not in supported_base_models:
        raise Exception(f"Invalid base model {base_model}")
    filename = ''
    if base_model == 'meta-llama/Llama-3.1-70B-Instruct':
        fileName = resource_filename('tensorkube', 'configurations/axolotl_configs/llama-3.1-70B-Instruct.yaml')
    elif base_model == 'meta-llama/Llama-3.1-8B-Instruct':
        fileName = resource_filename('tensorkube', 'configurations/axolotl_configs/llama-3.1-8B-Instruct.yaml')
    else:
        raise Exception(f"Invalid base model {base_model}")
    with open(fileName, 'r') as f:
        yaml_dict = yaml.safe_load(f)
    return yaml_dict


def create_fine_tuning_job(job_name: str, job_id: str, gpus: int, gpu_type: str, max_scale: int, base_model: str,
                           dataset: str, epochs: int, cpu: float = DEFAULT_CPU, memory: float = DEFAULT_MEMORY,
                           secrets: List[str] = [], micro_batch_size: Optional[int] = None,
                           learning_rate: Optional[float] = None, lora_r: Optional[int] = None,
                           val_set_size: Optional[float] = None,
                           wandb_entity: Optional[str] = None, hf_org_id: Optional[str] = None, **kwargs) -> None:
    # create the job definition
    env = 'keda'
    # validate the base model
    if base_model not in supported_base_models:
        raise Exception(f"Invalid base model {base_model}")
    try:
        generate_kubeconfig()
    except Exception as e:
        raise Exception(f"Failed to generate kubeconfig, {e}")
    try:
        exist, _ = is_existing_queued_job(job_name=job_name)
        if not exist:
            deploy_job(job_name=job_name, env=env, gpus=gpus, gpu_type=gpu_type, cpu=cpu, memory=memory,
                       max_scale=max_scale, image_tag=AXOLOTL_IMAGE_SCALED_JOB, secrets=secrets, job_type='axolotl')
    except ApiException as e:
        if e.status == 409:
            click.echo(f"Job {job_name} already exists.")
        else:
            raise Exception(f"Failed to deploy job: {e}")
    base_yaml_dict = get_axolotl_base_config_by_model(base_model=base_model)
    config = get_config_from_paramaters(base_model=base_model, dataset=dataset, epochs=epochs,
                                        micro_batch_size=micro_batch_size, learning_rate=learning_rate, lora_r=lora_r,
                                        wandb_entity=wandb_entity, wandb_project=job_name,
                                        val_set_size=val_set_size,
                                        wandb_name=job_id,
                                        base_config=base_yaml_dict,
                                        hf_org_id=hf_org_id)
    yaml_dict = {**base_yaml_dict, **config}
    yaml_dict = {**yaml_dict, **kwargs}
    payload = json.dumps(yaml_dict)
    queue_job(job_name, job_id, payload)
    return



from tensorkube.services.configure_service import update_cfn_configure_stack


updated_params = {
    "TemplatesVersion": "v0.0.2",
    "AWSAccessLambdaFunctionImageVersion": "v1.0.1",
    "CliVersion": "0.0.71",
    "EksAccessLambdaFunctionImageVersion": "v1.0.0"
}
update_cfn_configure_stack(updated_parameters=updated_params, test=False)




from tensorkube.services.build import configure_buildkit_irsa
from tensorkube.services.environment_service import list_environments
from tensorkube.services.ecr_service import patch_lifecycle_all_tensorkube_ecr_repositories
from tensorkube.services.job_queue_service import deploy_job
from tensorkube.services.k8s_service import get_jobs_by_container_image
from tensorkube.services.knative_service import add_runtimeclass_knative_features, get_gpu_type_from_instance_family
from tensorkube.services.nydus import install_nydus
from tensorkube.services.train import AXOLOTL_IMAGE_SCALED_JOB


def apply():
    try:
        install_nydus()
        add_runtimeclass_knative_features()
        patch_lifecycle_all_tensorkube_ecr_repositories()
        envs  = list_environments()
        for env in envs:
            configure_buildkit_irsa(env_name=env)
        # trying to migrate jobs
        env = 'keda'
        jobs = get_jobs_by_container_image(AXOLOTL_IMAGE_SCALED_JOB, namespace=env)
        for job in jobs:
            containers = job.get("spec", {}).get("jobTargetRef", {}).get("template", {}).get("spec", {}).get(
                "containers", [])
            node_selector = job.get("spec", {}).get("jobTargetRef", {}).get("template", {}).get("spec", {}).get(
                "nodeSelector", {})
            instance_family = node_selector.get("karpenter.k8s.aws/instance-family")
            gpu_type = get_gpu_type_from_instance_family(instance_family) if instance_family else None
            max_scale = job.get("spec", {}).get("maxReplicaCount")
            job_name = job["metadata"]["name"]

            # Extract secrets from volumes
            volumes = job.get("spec", {}).get("jobTargetRef", {}).get("template", {}).get("spec", {}).get("volumes", [])
            secrets = []
            for volume in volumes:
                if volume.get("name") == "secrets" and "projected" in volume:
                    for source in volume["projected"].get("sources", []):
                        if "secret" in source:
                            secrets.append(source["secret"]["name"])

            for container in containers:
                if container.get("image") == AXOLOTL_IMAGE_SCALED_JOB:
                    resources = container.get("resources", {})
                    gpus = resources.get("limits", {}).get("nvidia.com/gpu")
                    cpu = resources.get("requests", {}).get("cpu")
                    memory = resources.get("requests", {}).get("memory")

                    deploy_job(job_name=job_name, env=env, gpus=gpus, gpu_type=gpu_type, cpu=cpu, memory=memory,
                        max_scale=max_scale, image_tag=AXOLOTL_IMAGE_SCALED_JOB, secrets=secrets, job_type='axolotl',
                               update=True)
            print(f"Migrated the job {job_name} to the latest version")

    except Exception as e:
        raise e



import os
import platform
import shutil
import subprocess
import uuid
from typing import Optional

import boto3
import botocore
import click
import semver

from tensorkube.constants import CliColors, REGION
from tensorkube.constants import get_cluster_name, DEFAULT_NAMESPACE, LOCKED_AWS_CLI_VERSION
from tensorkube.services.error import CLIVersionError
from tensorkube.services.s3_service import list_s3_buckets


def get_session_region():
    return boto3.session.Session().region_name


def get_eks_client():
    return boto3.client("eks")


def get_cloudformation_client():
    return boto3.client("cloudformation")


def get_ec2_client():
    return boto3.client("ec2")


def get_iam_client():
    return boto3.client("iam")


def get_aws_account_id():
    sts = boto3.client('sts')
    identity = sts.get_caller_identity()
    return identity['Account']


def get_aws_user_arn() -> str:
    sts = boto3.client('sts')
    identity = sts.get_caller_identity()
    return identity['Arn']


def get_principal_arn_from_identity_center_arn(arn: str) -> str:
    arn_parts = arn.split(':')
    account_no = arn_parts[4]
    role_name = arn_parts[5].split('/')[1]
    principal_arn = f"arn:aws:iam::{account_no}:role/aws-reserved/sso.amazonaws.com/{role_name}"
    return principal_arn


def get_aws_user_principal_arn() -> str:
    sts = boto3.client('sts')
    identity = sts.get_caller_identity()
    if 'assumed-role' in identity['Arn']:
        return get_principal_arn_from_identity_center_arn(identity['Arn'])
    return identity['Arn']


def get_karpenter_namespace():
    return "kube-system"


def get_karpenter_version():
    return "0.37.0"


def get_aws_default_region():
    return "us-east-1"


def get_kubernetes_context_name():
    return f"{get_cluster_name()}.{get_aws_default_region()}.eksctl.io"


def get_credentials():
    return boto3.Session().get_credentials().get_frozen_credentials()


def are_credentials_valid(credentials):
    sts = boto3.client('sts', aws_access_key_id=credentials.access_key, aws_secret_access_key=credentials.secret_key,
                       aws_session_token=credentials.token)
    try:
        sts.get_caller_identity()
        return True
    except botocore.exceptions.ClientError as e:
        return False


# TODO!: make function generic to get any config value
def get_bucket_name(env_name: Optional[str] = None, type: str = "build"):
    buckets = list_s3_buckets()
    if env_name and env_name != DEFAULT_NAMESPACE:
        prefix = f'{get_cluster_name()}-{env_name}-{type}-bucket-'
    else:
        prefix = f'{get_cluster_name()}-{type}-bucket-'
    # checks if such a bucket already exists
    for bucket in buckets:
        if bucket['Name'].startswith(prefix):
            return bucket['Name']
    else:
        bucket_name = f"{prefix}{str(uuid.uuid4())[:18]}"
        if len(bucket_name) > 63:
            raise ValueError("Environment name is too long. Please use a shorter name.")
        return bucket_name


def get_existing_bucket_name(env_name: Optional[str] = None, type: str = "build"):
    buckets = list_s3_buckets()
    if env_name and env_name != DEFAULT_NAMESPACE:
        prefix = f'{get_cluster_name()}-{env_name}-{type}-bucket-'
    else:
        prefix = f'{get_cluster_name()}-{type}-bucket-'
    # checks if such a bucket already exists
    for bucket in buckets:
        if bucket['Name'].startswith(prefix):
            return bucket['Name']

    return None


def check_and_install_aws_cli():
    """Check if aws cli is installed and if not install it."""
    try:
        result = subprocess.run(["aws", "--version"], capture_output=True, text=True, check=True)

        # The version is in the format: "aws-cli/2.13.4 Python/3.9.6 Linux/x86_64"
        version_output = result.stdout.strip()
        # Extract the AWS CLI version using string manipulation
        aws_version = semver.VersionInfo.parse(version_output.split(" ")[0].split("/")[1])
        locked_cli_version = semver.VersionInfo.parse(LOCKED_AWS_CLI_VERSION)
        if (aws_version.major, aws_version.minor) < (locked_cli_version.major, locked_cli_version.minor):
            text = f"AWS CLI version is {aws_version}. Please upgrade AWS CLI to version above {LOCKED_AWS_CLI_VERSION}."
            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
            raise CLIVersionError(text)
    except Exception as e:
        if isinstance(e, CLIVersionError):
            raise e
        click.echo(
            click.style("aws-cli not found. Proceeding with installation. Might require sudo password.", bold=True,
                        fg=CliColors.WARNING.value))
        try:
            system = platform.system().lower()
            architecture = platform.machine()
            print(f"Detected operating system: {system.capitalize()}")
            print(f"Detected architecture: {architecture}")
            url = ""
            file_name = ""
            # Define the download URL and file name based on the OS
            if system == "linux":
                url = f"https://awscli.amazonaws.com/awscli-exe-linux-{architecture}-{LOCKED_AWS_CLI_VERSION}.zip"
                file_name = f"awscliv2-{LOCKED_AWS_CLI_VERSION}.zip"
            elif system == "darwin":
                url = f"https://awscli.amazonaws.com/AWSCLIV2-{LOCKED_AWS_CLI_VERSION}.pkg"
                file_name = f"AWSCLIV2-{LOCKED_AWS_CLI_VERSION}.pkg"
            else:
                print("Unsupported operating system. Please install AWS CLI manually.")
                raise Exception("Unsupported operating system.")
            # Download the installer
            print(f"Downloading AWS CLI installer from {url}...")
            subprocess.run(["curl", "-fsSL", url, "-o", file_name], check=True)
            print("AWS CLI installer downloaded successfully.")
            if system == "linux":
                # Unzip the installer
                print("Unzipping the AWS CLI installer...")
                subprocess.run(["unzip", "-o", file_name], check=True)
                print("AWS CLI installer unzipped successfully.")

                # Install the AWS CLI
                print("Installing AWS CLI...")
                subprocess.run(["sudo", "./aws/install"], check=True)
                print("AWS CLI installed successfully.")

            elif system == "darwin":
                # Use the macOS installer
                print("Installing AWS CLI using macOS package...")
                subprocess.run(["sudo", "installer", "-pkg", file_name, "-target", "/"], check=True)
                print("AWS CLI installed successfully.")
        except subprocess.CalledProcessError as e:
            print(f"Error occurred during installation: {e}")
        finally:
            if os.path.exists(file_name):
                os.remove(file_name)
            if os.path.exists("aws"):
                shutil.rmtree("aws")
            print("Clean-up completed.")

# https://github.com/eksctl-io/eksctl/blob/f4d4076ee22c9026a11a70434b802a7e0f61bd46/pkg/az/az.go#L24
# taken from eksctl source code
zoneIDsToAvoid = {
    "us-east-1": ["use1-az3"],
    "us-west-1": ["usw1-az2"],
    "ca-central-1": ["cac1-az3"],
    "cn-north-1": ["cnn1-az4"],
}


def get_eks_control_plane_available_zones_for_region(region:str=REGION):
    ec2 = boto3.client('ec2', region_name=region)
    response = ec2.describe_availability_zones()
    return [zone['ZoneName'] for zone in response['AvailabilityZones'] if zone['ZoneId'] not in zoneIDsToAvoid.get(region, []) and zone['ZoneType'] == "availability-zone"]
